--EnhancedChatFilter 内部定义
local ecfVersioin = GetAddOnMetadata("EnhancedChatFilter","Version")
local _, ecf = ...
local utf8replace = ecf.utf8replace -- utf8.lua
local L = ecf.L -- locales.lua
local chatLines = {}
local prevLineID = 0
local filterResult = nil
local filterCharList = "[|@!/<>\"`'_#&;:~\\]"
local filterCharListRegex = "[%(%)%.%%%+%-%*%?%[%]%$%^={}]"
local allowWisper = {}
local config

local gsub, select, ipairs, tremove, tinsert, pairs, strsub, getn, sort, format, tonumber, strfind, strlen, strmatch =
gsub, select, ipairs, tremove, tinsert, pairs, strsub, getn, sort, format, tonumber, string.find, strlen, strmatch

--信息中将被过滤器过滤的 UTF-8 符号（常被用于干扰）
local UTF8Symbols = {['·']='',['＠']='',['＃']='',['％']='',
	['＆']='',['＊']='',['——']='',['＋']='',['｜']='',['～']='',['　']='',
	['，']='',['。']='',['、']='',['？']='',['！']='',['：']='',['；']='',
	['’']='',['‘']='',['“']='',['”']='',['【']='',['】']='',['『']='',
	['』']='',['《']='',['》']='',['（']='',['）']='',['￥']='',['＝']='',
	['…']='',['……']='',['１']='1',['２']='2',['３']='3',['４']='4',['５']='5',
	['６']='6',['７']='7',['８']='8',['９']='9',['０']='0',['⒈']='1',['⒉']='2',
	['⒊']='3',['⒋']='4',['⒌']='5',['⒍']='6',['⒎']='7',['⒏']='8',['⒐']='9',
	['〔']='',['〕']='',['〈']='',['〉']='',['‖']=''}

--http://www.wowwiki.com/USERAPI_StringHash
local function StringHash(text)
	local counter = 1
	local len = string.len(text)
	for i = 1, len, 3 do
	counter = math.fmod(counter*8161, 4294967279) +  -- 2^32 - 17: Prime!
		(string.byte(text,i)*16776193) +
		((string.byte(text,i+1) or (len-i+256))*8372226) +
		((string.byte(text,i+2) or (len-i+256))*3932164)
	end
	return math.fmod(counter, 4294967291) -- 2^32 - 5: Prime (and different from the prime in the loop)
end

local EnhancedChatFilter = LibStub("AceAddon-3.0"):NewAddon("EnhancedChatFilter", "AceConsole-3.0", "AceEvent-3.0")

--数据库初始设置
local defaults = {
	profile = {
		enableFilter = true, --ECF聊天过滤总开关
		enableWisper = false, --密语白名单过滤
		enableDND = true, --“忙碌”玩家过滤
		enableRPT = true, --重复聊天过滤
		enableCFA = true, --成就刷屏过滤
		enableRAF = false, --RaidAlert刷屏过滤
		enableQRF = false, --QuestReport刷屏过滤
		enableIGM = false, --扩展屏蔽名单
		blackWordList = {}, --关键词列表
		ignoreMoreList = {}, --扩展屏蔽名单
		minimap = {
			hide = false, --小地图图标选项
		},
		chatLinesLimit = 20,
		stringDifferenceLimit = 0.1,
		debugMode = false,
	}
}

--旧关键词表转化成新的
local function convert()
	if(config.blackList) then
		for _,v in ipairs(config.blackList) do
			config.blackWordList[v[1]] = v[2] or true
		end
	end
	config.blackList = nil
end

--创建小地图图标数据
local ecfLDB = LibStub("LibDataBroker-1.1"):NewDataObject("Enhanced Chat Filter", {
	type = "data source",
	text = "Enhanced Chat Filter",
	icon = "Interface\\Icons\\Trade_Archaeology_Orc_BloodText",
	OnClick = function() EnhancedChatFilter:EnhancedChatFilterOpen() end,
	OnTooltipShow = function(tooltip)
		tooltip:AddLine("|cffecf0f1Enhanced Chat Filter|r\n"..L["ClickToOpenConfig"])
	end
})

--创建 Libstub 变量
local icon = LibStub("LibDBIcon-1.0")

--初始化插件
function EnhancedChatFilter:OnInitialize()
	EnhancedChatFilter:RegisterChatCommand("ecf", "EnhancedChatFilterOpen")
	EnhancedChatFilter:RegisterChatCommand("ecf-clear", "EnhancedChatFilterClear")
	EnhancedChatFilter:RegisterChatCommand("ecf-iml", "EnhancedChatFilterIgnoreMoreList")
	EnhancedChatFilter:RegisterChatCommand("ecf-cleariml", "EnhancedChatFilterClearIMList")
	EnhancedChatFilter:RegisterChatCommand("ecf-debug", "EnhancedChatFilterDebug")

	self.db = LibStub("AceDB-3.0"):New("ecfDB", defaults, "Default")
	config = self.db.profile
	icon:Register("Enhanced Chat Filter", ecfLDB, config.minimap)
	convert()
end

local function checkBlacklist(blackWord, typeModus) -- 确保关键词自身不会被预处理干扰
	local newWord = blackWord:gsub("%s", ""):gsub(filterCharList, "")
	if (typeModus ~= "regex") then newWord=newWord:gsub(filterCharListRegex, "") end
	newWord = utf8replace(newWord, UTF8Symbols)
	if(newWord ~= blackWord) then return true end
end

--Options
--这些设置不会被保存
local scrollHighlight = {}
local regexToggle = nil
local stringIO = ""

local options = {
	type = "group",
	name = "EnhancedChatFilter",
	get = function(info) return config[info[#info]] end,
	set = function(info, value) config[info[#info]] = value end,
	args = {
		ToggleTab = {
			type = "group",
			name = L["General"],
			order = 1,
			args = {
				enableFilter = {
					type = "toggle",
					name = L["MainFilter"],
					order = 1
				},
				MinimapToggle = {
					type = "toggle",
					name = L["MinimapButton"],
					get = function() return config.minimap.hide end,
					set = function(_,toggle)
							config.minimap.hide = toggle
							if(toggle) then icon:Show("Enhanced Chat Filter") else icon:Hide("Enhanced Chat Filter") end
						end,
					order = 2
				},
				line1 = {
					type = "header",
					name = L["Filters"],
					order = 10,
				},
				enableDND = {
					type = "toggle",
					name = L["DND"],
					desc = L["DNDfilterTooltip"],
					order = 11
				},
				enableRPT = {
					type = "toggle",
					name = L["Repeat"],
					desc = L["RepeatFilterTooltip"],
					order = 12
				},
				enableCFA = {
					type = "toggle",
					name = L["Achievement"],
					desc = L["AchievementFilterTooltip"],
					order = 13
				},
				enableRAF = {
					type = "toggle",
					name = L["RaidAlert"],
					desc = L["RaidAlertFilterTooltip"],
					order = 14
				},
				enableQRF = {
					type = "toggle",
					name = L["QuestReport"],
					desc = L["QuestReportFilterTooltip"],
					order = 15
				},
				enableIGM = {
					type = "toggle",
					name = L["IgnoreMoreList"],
					order = 16
				},
				line2 = {
					type = "header",
					name = L["UseWithCare"],
					order = 20,
				},
				enableWisper = {
					type = "toggle",
					name = L["WhisperWhitelistMode"],
					desc = L["WhisperWhitelistModeTooltip"],
					order = 21
				},
			},
		},
		blackListTab = {
			type = "group",
			name = L["BlackworldList"],
			order = 2,
			args = {
				blackword = {
					type = "input",
					name = L["AddBlackWordTitle"],
					order = 1,
					get = nil,
					set = function(_,value)
						if (checkBlacklist(value, regexToggle) == true) then
							print(value..L["IncludeAutofilteredWord"])
						else
							config.blackWordList[value] = (regexToggle and "regex") or true
							scrollHighlight = {}
						end
					end,
				},
				regex = {
					type = "toggle",
					name = L["Regex"],
					desc = L["RegexTooltip"],
					order = 2,
					get = function() return regexToggle end,
					set = function(_,value) regexToggle = value end,
				},
				DeleteButton = {
					type = "execute",
					name = L["Remove"],
					order = 3,
					func = function()
						for key in pairs(scrollHighlight) do config.blackWordList[key] = nil end
						scrollHighlight = {}
					end,
				},
				ClearUpButton = {
					type = "execute",
					name = L["ClearUp"],
					order = 4,
					func = function() EnhancedChatFilter:EnhancedChatFilterClear() end,
					confirm = true,
					confirmText = L["DoYouWantToClearBlackList?"],
				},
				blackWordList = {
					type = "multiselect",
					name = L["BlackworldList"],
					order = 10,
					get = function(_,key) return scrollHighlight[key] end,
					set = function(_,key,value) scrollHighlight[key] = value or nil end,
					values = function()
						local blacklistname = {}
						for key in pairs(config.blackWordList) do blacklistname[key] = key end
						return blacklistname
					end,
				},
				line1 = {
					type = "header",
					name = L["StringIO"],
					order = 20,
				},
				stringconfig = {
					type = "input",
					name = "",
					order = 21,
					get = function() return stringIO end,
					set = function(_,value) stringIO = value end,
					width = "double",
				},
				import = {
					type = "execute",
					name = L["Import"],
					order = 22,
					func = function()
						local wordString, HashString = strsplit("@", stringIO)
						if (tonumber(HashString) ~= tonumber(StringHash(wordString))) then
							print(L["StringHashMismatch"])
							return
						end
						local newBlackList = {strsplit(";", wordString)}
						for _, blacklist in ipairs(newBlackList) do
							if (blacklist ~= nil) then
								local imNewWord, imTypeWord = strsplit(",",blacklist)
								if (checkBlacklist(imNewWord, imTypeWord) == true) then
									print(imNewWord..L["IncludeAutofilteredWord"])
								else
									config.blackWordList[imNewWord] = imTypeWord or true
								end
							end
						end
						stringIO = ""
						print(L["ImportSucceeded"])
					end,
				},
				export = {
					type = "execute",
					name = L["Export"],
					order = 23,
					func = function()
						local blackStringList = {}
						for key,v in pairs(config.blackWordList) do
							if (checkBlacklist(key, v) == true) then
								print(key..L["IncludeAutofilteredWord"])
							else
								if (v == true) then
									blackStringList[#blackStringList+1] = key
								else
									blackStringList[#blackStringList+1] = key..","..v
								end
							end
						end
						local blackString = table.concat(blackStringList,";")
						stringIO = blackString.."@"..StringHash(blackString)
					end,
				},
			},
		},
		Advanced = {
			type = "group",
			name = L["AdvancedConfig"],
			order = 4,
			args = {
				chatLinesLimit = {
					type = "range",
					name = L["chatLinesLimitSlider"],
					desc = L["chatLinesLimitSliderTooltips"],
					order = 1,
					min = 1,
					max = 100,
					step = 1,
					bigStep = 5,
				},
				stringDifferenceLimit = {
					type = "range",
					name = L["stringDifferenceLimitSlider"],
					desc = L["stringDifferenceLimitSliderTooltips"],
					order = 2,
					min = 0,
					max = 1,
					step = 0.01,
					bigStep = 0.1,
					isPercent = true,
				},
				line1 = {
					type = "header",
					name = "",
					order = 10,
				},
				debugMode = {
					type = "toggle",
					name = "DebugMode",
					desc = "For test only",
					order = 11,
				},
			},
		},
	},
}
LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("EnhancedChatFilter", options)
LibStub("AceConfigDialog-3.0"):AddToBlizOptions("EnhancedChatFilter", "EnhancedChatFilter")

--扩展黑名单
local function SendMessage(event, msg)
	local info = ChatTypeInfo[strsub(event, 10)]
	for i = 1, NUM_CHAT_WINDOWS do
		local ChatFrames = _G["ChatFrame"..i]
		if (ChatFrames and ChatFrames:IsEventRegistered(event)) then
			ChatFrames:AddMessage(msg, info.r, info.g, info.b)
		end
	end
end

local function ignoreMore(player)
	if (config.enableIGM ~= true) then return end
	if (not player) then return end
	local ignore = nil
	if GetNumIgnores() >= 50 then
		for i = 1, GetNumIgnores() do
			local name = GetIgnoreName(i)
			if (player == name) then
				ignore = true
				break
			end
		end
		if (not ignore) then
			local trimmedPlayer = Ambiguate(player, "none")
			tinsert(config.ignoreMoreList,{trimmedPlayer})
			if config.debugMode then print("Added to ECF ignoreMoreList!") end
			SendMessage("CHAT_MSG_SYSTEM", format(ERR_IGNORE_ADDED_S, trimmedPlayer))
		end
	end
end

hooksecurefunc("AddIgnore", ignoreMore)
hooksecurefunc("AddOrDelIgnore", ignoreMore)

--禁用暴雪内部语言过滤器
local GetCVar,SetCVar = GetCVar,SetCVar

local profanityFilter=CreateFrame("Frame")

local function allowSwearing()
	if GetCVar("profanityFilter")~="0" then SetCVar("profanityFilter", "0") end
end

profanityFilter:SetScript("OnEvent", allowSwearing)
profanityFilter:RegisterEvent("VARIABLES_LOADED")
profanityFilter:RegisterEvent("CVAR_UPDATE")
profanityFilter:RegisterEvent("PLAYER_ENTERING_WORLD")
profanityFilter:RegisterEvent("BN_MATURE_LANGUAGE_FILTER")
profanityFilter:RegisterEvent("BN_CONNECTED")

allowSwearing() -- 确认再次载入

--登陆/好友列表更新时同步更新允许密语玩家列表
local login = nil
local ecfFrame = CreateFrame("Frame")
ecfFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
ecfFrame:RegisterEvent("FRIENDLIST_UPDATE")
ecfFrame:SetScript("OnEvent", function(self, event)
	if event == "PLAYER_ENTERING_WORLD" then
		ShowFriends() --强制刷新一次好友列表
	else
		if not login then --登陆时只添加一次
			login = true
			local num = GetNumFriends()
			for i = 1, num do
				local n = GetFriendInfo(i)
				--添加好友到允许名单
				if n then allowWisper[n] = true end
			end
			return
		end
	end
	if config.debugMode then for k in pairs(allowWisper) do print("ECF allowed: "..k) end end
end)

--自己主动密语的玩家加入安全名单
local function addToAllowWisper(self,_,_,player)
	local trimmedPlayer = Ambiguate(player, "none")
	allowWisper[trimmedPlayer] = true
end

--字符串不同程度, 范围0..1, 0为完全相同, 1为完全不同
local function stringDifference(stringA, stringB)
	local len_a, len_b = #stringA, #stringB
	local templast, temp = {}, {}
	for j=0, len_b do templast[j+1] = j end
	for i=1, len_a do
		temp[1] = i
		for j=1, len_b do
			if(stringA:sub(i,i) == stringB:sub(j,j)) then
				temp[j+1] = templast[j]
			else
				temp[j+1] = min(templast[j+1], temp[j], templast[j]) + 1
			end
		end
		for j=0, len_b do templast[j+1]=temp[j+1] end
	end
	return temp[len_b+1]/(len_a+len_b)
end

-------------------------------------- Filters ------------------------------------

local chatChannel = {["CHAT_MSG_WHISPER"] = 1, ["CHAT_MSG_SAY"] = 2, ["CHAT_MSG_CHANNEL"] = 3, ["CHAT_MSG_YELL"] = 3, ["CHAT_MSG_PARTY"] = 4, ["CHAT_MSG_PARTY_LEADER"] = 4, ["CHAT_MSG_RAID"] = 4, ["CHAT_MSG_RAID_LEADER"] = 4, ["CHAT_MSG_RAID_WARNING"] = 4, ["CHAT_MSG_INSTANCE_CHAT"] = 4,["CHAT_MSG_INSTANCE_CHAT_LEADER"] = 4, ["CHAT_MSG_DND"] = 101}

local function ECFfilter(self,event,msg,player,_,_,_,flags,_,_,_,_,lineID)
	--如果不启用过滤直接返回
	if(config.enableFilter ~= true) then return end

	--避免自己被过滤
	local trimmedPlayer = Ambiguate(player, "none")
	if UnitIsUnit(trimmedPlayer,"player") then return end

	--对 GM 和 DEV 的密语不做处理
	if type(flags) == "string" and (flags == "GM" or flags == "DEV") then return end

	--如果这行被处理过，则直接引用之前的结果，避免多聊天窗口时触发重复过滤器
	if(lineID == prevLineID) then
		return filterResult
	else
		prevLineID = lineID
		filterResult = nil
	end
	
	if config.debugMode then print("RAWMsg: "..msg) end

	if(config.enableIGM == true and chatChannel[event] <= 1) then --拓展黑名单过滤
		for _,ignorePlayer in ipairs(config.ignoreMoreList) do
			if (trimmedPlayer == ignorePlayer[1]) then
				if config.debugMode then print("Trigger: IgnoreMore Filter") end
				filterResult = true
				return true
			end
		end
	end

	if(config.enableWisper == true and chatChannel[event] <= 1) then --密语白名单过滤
		--强制刷新一次好友列表
		ShowFriends()
		--同工会，同团队，同小队和好友不做处理
		if allowWisper[trimmedPlayer] or UnitIsInMyGuild(trimmedPlayer) or UnitInRaid(trimmedPlayer) or UnitInParty(trimmedPlayer) then return end
		--战网好友不做处理
		for i = 1, select(2, BNGetNumFriends()) do
			local GameAccount = BNGetNumFriendGameAccounts(i)
			for j = 1, GameAccount do
				local _, rName, rGame = BNGetFriendGameAccountInfo(i, j)
				if (rName == trimmedPlayer and rGame == "WoW") then return end
			end
		end
		if config.debugMode then print("Trigger: WhiteListMode") end
		filterResult = true
		return true
	end

	if(config.enableDND == true and (chatChannel[event] <= 3 or chatChannel[event] == 101)) then --"忙碌"玩家过滤
		if ((type(flags) == "string" and flags == "DND") or chatChannel[event] == 101) then
			if config.debugMode then print("Trigger: DND Filter") end
			filterResult = true
			return true
		end
	end

	if(chatChannel[event] <= 3) then --关键词过滤
		--删除颜色代码, 物品链接标志, 空格以及一些符号
		local filterString = msg:upper():gsub("|C[0-9A-F]+",""):gsub("|H[^|]+|H",""):gsub("|H|R",""):gsub("%s", ""):gsub(filterCharList, "")

		--删除 UTF-8 干扰符
		filterString = utf8replace(filterString, UTF8Symbols)
		local newfilterString = filterString:gsub(filterCharListRegex, "")

		--从处理过的聊天信息中过滤包含黑名单词语的聊天内容
		for keyWord,v in pairs(config.blackWordList) do
			local currentString
			if (v ~= "regex") then -- 如果不是正则，也过滤正则中使用的特殊字符
				keyWord = keyWord:upper()
				currentString = newfilterString
			else
				currentString = filterString
			end
			--检查常规黑名单
			if (strfind(currentString,keyWord)) then
				if config.debugMode then print("Trigger: Keyword: "..keyWord) end
				filterResult = true
				return true --deletion of message from chat
			end
		end
	end

	local RaidAlertTagList = {"%*%*(.+)%*%*", "EUI:(.+)施放了", "EUI:(.+)中断", "EUI:(.+)就绪"}  --RaidAlert 特征
	local QuestReportTagList = {"任务进度提示%s?[:：]", "%(任务完成%)", "<大脚组队提示>", "%[接受任务%]", "<大脚团队提示>"} --QuestReport 特征

	if (config.enableRAF == true and IsInGroup() and (chatChannel[event] == 4 or chatChannel[event] == 2)) then
		for _,RaidAlertTag in ipairs(RaidAlertTagList) do
			if(msg:find(RaidAlertTag)) then
				if config.debugMode then print("Trigger: "..RaidAlertTag.." in RaidAlertTag") end
				filterResult = true
				return true
			end
		end
	end

	if (config.enableQRF == true and IsInGroup() and (chatChannel[event] == 4 or chatChannel[event] == 2)) then
		for _,QuestReportTag in ipairs(QuestReportTagList) do
			if(msg:find(QuestReportTag)) then
				if config.debugMode then print("Trigger: "..QuestReportTag.." in QuestReportTag") end
				filterResult = true
				return true
			end
		end
	end

	if(config.enableRPT == true and chatChannel[event] <= 4) then --重复信息过滤
		--删除掉表情和所有标点符号, 但如果对话只有这些内容则保留
		local msgLine = msg:gsub("%p+", ""):gsub("{.*}","")
		if(msgLine == "") then msgLine = msg end

		--处理重复信息
		local msgtable = {Sender = trimmedPlayer, Msg = msgLine, Time = GetTime()}
		for i=1, #chatLines do
			--如果一个人在0.6秒钟之内发了多条信息, 过滤
			--如果一个人的信息完全相同, 过滤
			if (chatLines[i].Sender == msgtable.Sender and ((msgtable.Time - chatLines[i].Time) < 0.600 or stringDifference(chatLines[i].Msg,msgtable.Msg) <= config.stringDifferenceLimit)) then
				chatLines[i] = msgtable
				if config.debugMode then print("Trigger: Repeat Filter") end
				filterResult = true
				return true
			end
			if i >= config.chatLinesLimit then tremove(chatLines, 1) end
		end
		tinsert(chatLines, msgtable)
	end
end

ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", addToAllowWisper)
ChatFrame_AddMessageEventFilter("CHAT_MSG_DND", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_SAY", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_CHANNEL", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_YELL", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_PARTY", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_PARTY_LEADER", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID_LEADER", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID_WARNING", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_INSTANCE_CHAT", ECFfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_INSTANCE_CHAT_LEADER", ECFfilter)

-- create startup event when player logs in
local startUp = CreateFrame("Frame")
startUp:RegisterEvent("PLAYER_LOGIN")

--function that gets called when an event within startUp triggers.
startUp:SetScript("OnEvent", function(self)
	print("|cffffff00Enhanced Chat Filter MOD|r|cff3498db v"..ecfVersioin.."|cffecf0f1 - "..L["CommandLine"].." '/ecf'")
	if(config.minimap.hide) then icon:Show("Enhanced Chat Filter") else icon:Hide("Enhanced Chat Filter") end
end)

--method run on /ecf
function EnhancedChatFilter:EnhancedChatFilterOpen()
	if(InCombatLockdown()) then return end
	InterfaceOptionsFrame_OpenToCategory("EnhancedChatFilter")
end

--method run on /ecf-clear
function EnhancedChatFilter:EnhancedChatFilterClear()
	config.blackWordList = {}
	scrollHighlight = {}
	print(L["ClearedBlackWordList"])
end

--method run on /ecf-iml
function EnhancedChatFilter:EnhancedChatFilterIgnoreMoreList()
	local ignoreMoreList = config.ignoreMoreList
	print(L["IgnoreMoreList"]..": ")
	for _,ignorePlayer in ipairs(ignoreMoreList) do
		print(ignorePlayer[1])
	end
end

--method run on /ecf-cleariml
function EnhancedChatFilter:EnhancedChatFilterClearIMList()
	EnhancedChatFilter:ClearIgnoreMoreList()
	print(L["ClearedIgnoreMoreList"])
end

--method run on /ecf-debug
function EnhancedChatFilter:EnhancedChatFilterDebug()
	if(config.debugMode ~= nil) then
		config.debugMode=nil
		print("Debug Mode Off!")
	else
		config.debugMode=true
		print("Debug Mode On!")
	end
end

-----------------------------------------------------------------------
-- ChatFilter 成就刷屏过滤模块
-----------------------------------------------------------------------

local ChatFilter = CreateFrame("Frame")
local achievements, alreadySent = {}, {}
local spamCategories, specialFilters = {[95] = true, [155] = true, [168] = true, [14807] = true, [15165] = true}, {[456] = true, [1400] = true, [1402] = true, [2186] = true, [2187] = true, [2903] = true, [2904] = true, [3004] = true, [3005] = true, [3117] = true, [3259] = true, [3316] = true, [3808] = true, [3809] = true, [3810] = true, [3817] = true, [3818] = true, [3819] = true, [4078] = true, [4079] = true, [4080] = true, [4156] = true, [4576] = true, [4626] = true, [5313] = true, [6954] = true, [7485] = true, [7486] = true, [7487] = true, [8089] = true, [8238] = true, [8246] = true, [8248] = true, [8249] = true, [8260] = true, [8306] = true, [8307] = true, [8398] = true, [8399] = true, [8400] = true, [8401] = true, [8430] = true, [8431] = true, [8432] = true, [8433] = true, [8434] = true, [8435] = true, [8436] = true, [8437] = true, [8438] = true, [8439] = true}

local function SendAchievement(event, achievementID, players)
	if (not players) then return end
	for k in pairs(alreadySent) do alreadySent[k] = nil end
	for i = getn(players), 1, -1 do
		if (alreadySent[players[i].name]) then
			tremove(players, i)
		else
			alreadySent[players[i].name] = true
		end
	end
	if (getn(players) > 1) then
		sort(players, function(a, b) return a.name < b.name end)
	end
	for i = 1, getn(players) do
		local class, color, r, g, b
		if (players[i].guid and players[i].guid:find("Player")) then
			class = select(2, GetPlayerInfoByGUID(players[i].guid))
			color = (CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS)[class]
		end
		if (not color) then
			local info = ChatTypeInfo[strsub(event, 10)]
			r, g, b = info.r, info.g, info.b
		else
			r, g, b = color.r, color.g, color.b
		end
		players[i] = format("|cff%02x%02x%02x|Hplayer:%s|h%s|h|r", r*255, g*255, b*255, players[i].name, players[i].name)
	end
	SendMessage(event, format("[%s]"..L["GotAchievement"].."%s!", table.concat(players, L["And"]), GetAchievementLink(achievementID)))
end

local function achievementReady(id, achievement)
	if (achievement.area and achievement.guild) then
		local playerGuild = GetGuildInfo("player")
		for i = getn(achievement.area), 1, -1 do
			local player = achievement.area[i].name
			if (UnitExists(player) and playerGuild and playerGuild == GetGuildInfo(player)) then
				tinsert(achievement.guild, tremove(achievement.area, i))
			end
		end
	end
	if (achievement.area and getn(achievement.area) > 0) then
		SendAchievement("CHAT_MSG_ACHIEVEMENT", id, achievement.area)
	end
	if (achievement.guild and getn(achievement.guild) > 0) then
		SendAchievement("CHAT_MSG_GUILD_ACHIEVEMENT", id, achievement.guild)
	end
end

local function ChatFrames_OnUpdate(self)
	local found
	for id, achievement in pairs(achievements) do
		if (achievement.timeout <= GetTime()) then
			achievementReady(id, achievement)
			achievements[id] = nil
		end
		found = true
	end
	if (not found) then
		self:SetScript("OnUpdate", nil)
	end
end

local function queueAchievementSpam(event, achievementID, playerdata)
	achievements[achievementID] = achievements[achievementID] or {timeout = GetTime() + 0.5}
	achievements[achievementID][event] = achievements[achievementID][event] or {}
	tinsert(achievements[achievementID][event], playerdata)
	ChatFilter:SetScript("OnUpdate", ChatFrames_OnUpdate)
end

local function ChatFilter_Achievement(self, event, msg, _, _, _, _, _, _, _, _, _, _, guid)
	if config.enableCFA ~= true then return end
	local achievementID = strmatch(msg, "achievement:(%d+)")
	if (not achievementID) then return end
	achievementID = tonumber(achievementID)
	local Name, Server
	if (guid and guid:find("Player")) then
		Name = select(6, GetPlayerInfoByGUID(guid))
		Server = select(7, GetPlayerInfoByGUID(guid))
		if (Name and Server and strlen(Server) > 0 and Server ~= GetRealmName()) then
			Name = Name.."-"..Server
		end
	end
	local playerdata = {name = Name, guid = guid}
	local categoryID = GetAchievementCategory(achievementID)
	if (spamCategories[categoryID] or spamCategories[select(2, GetCategoryInfo(categoryID))] or specialFilters[achievementID]) then
		queueAchievementSpam((event == "CHAT_MSG_GUILD_ACHIEVEMENT" and "guild" or "area"), achievementID, playerdata)
		return true
	end
end

ChatFrame_AddMessageEventFilter("CHAT_MSG_ACHIEVEMENT", ChatFilter_Achievement)
ChatFrame_AddMessageEventFilter("CHAT_MSG_GUILD_ACHIEVEMENT", ChatFilter_Achievement)

--拾取物品过滤器(阿什兰破碎骨骼+暗月马戏团)
local lootBadWords = {118043, 944, 71096}
local function lootfilter(self,_,msg)
	for _, itemId in ipairs(lootBadWords) do
		if (itemId == tonumber(strmatch(msg, "|Hitem:(%d+)"))) then
			return true
		end
	end
end

ChatFrame_AddMessageEventFilter("CHAT_MSG_LOOT", lootfilter)
